(* -------------------------------------------------------------*)
(*		ElabObj.ML	 				*)
(*		Derived Rules for Semantic Objects		*)
(* ------------------------------------------------------------ *)

structure elab_obj_lemmas =
struct
local open SML_Syntax elab_obj 
in

val VE0_full_def = rewrite_rule [pr_type_def] VE0_def;

(* -------------------------------------------------------------------- *)
(*	 	Meta-Level Rewrite Rules  				*)
(*	(used only to derive rules in this file)			*)
(* -------------------------------------------------------------------- *)

val defs =  
       [int_def, bool_def, exn_def, unit_def, pr_type_def,
	tyvars_tyvar, tyvars_funtype, tyvars_rectype_empty,
	tyvars_rectype, tyvars_constype_empty, tyvars_constype,
	tyvars_typescheme, tyvars_typefcn,
        Tyvars_empty, Tyvars_varenv_conenv, 
	Tyvars_tyenv, Tyvars_tuple]; 

(* -------------------------------------------------------------------- *)
(*	Derived Rules for Semantic Objects 				*)
(* -------------------------------------------------------------------- *)

(* type assertions for components of a context *)

fun type_tacfun prems = [cut_facts_tac prems 1 THEN
               REPEAT (eresolve_tac [SigmaE2] 1) THEN
               assume_tac 1];

val SE_type = rewrite_rule [StrEnv_def] 
  (prove_goalw thy [Context_def, Env_def] 
    "<T,U,SE,TE,VE,EE> : Context ==> SE : StrEnv" type_tacfun);

val TE_type = rewrite_rule [TyEnv_def] 
  (prove_goalw thy [Context_def, Env_def] 
    "<T,U,SE,TE,VE,EE> : Context ==> TE : TyEnv" type_tacfun);

val VE_type = rewrite_rule [VarEnv_def] 
  (prove_goalw thy [Context_def, Env_def] 
    "<T,U,SE,TE,VE,EE> : Context ==> VE : VarEnv" type_tacfun);

val EE_type = rewrite_rule [ExConEnv_def] 
  (prove_goalw thy [Context_def, Env_def] 
    "<T,U,SE,TE,VE,EE> : Context ==> EE : ExConEnv" type_tacfun);


fun tacfun prems = [cut_facts_tac prems 1 THEN fast_tac ZF_cs 1];

val Var_type = prove_goal thy 
 "x : Var ==> <Var, x> : (SUM x:{Var,Con,ExCon}. x)" tacfun;

val Con_type = prove_goal thy 
 "x : Con ==> <Con, x> : (SUM x:{Var,Con,ExCon}. x)" tacfun;

val ExCon_type = prove_goal thy 
 "x : ExCon ==> <ExCon, x> : (SUM x:{Var,Con,ExCon}. x)" tacfun;

val  [SE0_type, TE0_type, VE0_type, EE0_type] =
    ([rewrite_rule [C0_def] C0_type] RL 
     [SE_type, TE_type, VE_type, EE_type]) RL [mapD];

val set_rules =
   [singletonI, empty_Diff, Diff_0, Diff_cancel, 
    Un_0, empty_Un, Un_Diff1, Un_absorb, 
    Un_cons_equality_absorb,   Un_cons_equality];

fun tacfun prems = 
	     [cut_facts_tac prems 1, 
              rtac apply_equality1 1,
              REPEAT (ares_tac [SE0_type, TE0_type, VE0_type, EE0_type] 1)];

val apply_SE0 = prove_goal thy "<arg, s> : SE0 ==> SE0`arg = s" tacfun ;

val apply_TE0 = prove_goal thy "<arg, s> : TE0 ==> TE0`arg = s" tacfun ;

val apply_VE0 = prove_goal thy "<arg, s> : VE0 ==> VE0`arg = s" tacfun ;

val apply_EE0 = prove_goal thy "<arg, s> : EE0 ==> EE0`arg = s" tacfun ;

fun mk_mems def triv_thm =
let fun f th  = 
       (consI1 RS th) ::  f (consI2 RS th) handle _ => []
in map (rewrite_rule (map symmetric [def])) 
      (f (rewrite_rule [def] triv_thm))
end;

val VE0_mems = mk_mems VE0_full_def 
  (prove_goal thy "p : VE0 ==> p : VE0" 
            (fn prems => [cut_facts_tac prems 1, atac 1]));


val TE0_mems = mk_mems TE0_def 
  (prove_goal thy "p : TE0 ==> p : TE0" 
            (fn prems => [cut_facts_tac prems 1, atac 1]));

fun tacfun [] = [asm_simp_tac (ZF_ss addsimps set_rules) 1,
                 TRY (fast_tac (ZF_cs addIs [equalityI]) 1)];

val Tyvars_C0 = prove_goalw thy 
 ([T0_def, U0_def, C0_def, SE0_def, TE0_def, VE0_full_def, EE0_def] @ defs)
 "(Tyvars C0) = 0" tacfun;

val Tyvars_T0 = prove_goalw thy ([T0_def] @ defs) "(Tyvars T0) = 0" tacfun;

val Tyvars_U0 = prove_goalw thy ([U0_def] @ defs) "(Tyvars U0) = 0" tacfun;

val Tyvars_SE0 = prove_goalw thy ([SE0_def] @ defs) "(Tyvars SE0) = 0" tacfun;

val Tyvars_TE0 = prove_goalw thy ([TE0_def] @ defs) "(Tyvars TE0) = 0" tacfun;

val Tyvars_VE0 = prove_goalw thy ([VE0_full_def] @ defs) "(Tyvars VE0) = 0" tacfun;

val Tyvars_EE0 = prove_goalw thy ([EE0_def] @ defs) "(Tyvars EE0) = 0" tacfun;


val env_combine_equality = prove_goalw thy [env_combine_def]
 "[| SE @@ SE' = SE''; TE @@ TE' = TE''; \
\    VE @@ VE' = VE''; EE @@ EE' = EE'' ; \
\    <SE'', TE'', VE'', EE''> = E |] ==>\
\    <SE, TE, VE, EE> # <SE',TE', VE', EE'> = E"
  (fn prems => [cut_facts_tac prems 1, asm_simp_tac ZF_ss 1]);

val context_combine_equality = prove_goalw thy [context_combine_def]
 "[| T Un T' = T''; U Un U' = U''; E # E' = E'' ; \
\    <T'', U'', E''> = C |] ==>\
\    <T,U,E> ## <T',U',E'> = C"
  (fn prems => [cut_facts_tac prems 1, asm_simp_tac ZF_ss 1]);

val tyvars_int = 
 prove_goalw thy defs "tyvars int = 0" (fn prems => [rtac refl 1]);

val tyvars_bool = 
 prove_goalw thy defs "tyvars bool = 0" (fn prems => [rtac refl 1]);

val C0_combine_equality = 
prove_goalw thy [context_combine_def, env_combine_def, C0_def]
 "[| T0 Un T' = T''; U0 Un U' = U''; SE0 @@ SE' = SE'' ;   \
\     TE0 @@ TE' = TE''; VE0 @@ VE' = VE''; EE0 @@ EE' = EE'' ; \
\    <T'', U'', SE'', TE'', VE'', EE''> = C |] ==> \
\    C0 ## <T', U', SE', TE', VE', EE'> = C" 
(fn prems => [cut_facts_tac prems 1, asm_simp_tac ZF_ss 1]);

(* -------------------------------------------------------------------- *)
(*	 	Meta-Level Rewrite Rules  				*)
(*		(actually used in proofs)				*)
(* -------------------------------------------------------------------- *)

val meta_defs = 
 [close, Close, Close_empty, 
  C0_def, T0_def, U0_def, SE0_def, TE0_def, VE0_full_def, EE0_def,
  context_combine_def, env_combine_def, 
  tyvars_tyvar, tyvars_funtype, tyvars_rectype_empty,
  tyvars_rectype, tyvars_constype_empty, tyvars_constype,
  tyvars_typescheme, tyvars_typefcn, Tyvars_empty, Tyvars_tyenv, Tyvars_varenv_conenv, Tyvars_tuple, Tyvars_overwrite_init, pr_type_def]; 

fun tacfun prems = [cut_facts_tac prems 1, 
 asm_simp_tac (ZF_ss addsimps set_rules) 1];


val tyvars_tyvar_rule
  = prove_goalw thy meta_defs
	  "({a} = A) ==> tyvars <TyVar, a> = A" tacfun;

val tyvars_funtype_rule
  = prove_goalw thy meta_defs
	  "[| tyvars t = A; tyvars t' = A'; A Un A' = A''|] ==> \
 \    tyvars <FunType , t, t'> = A''" tacfun;

val tyvars_rectype_empty_rule
  = prove_goalw thy meta_defs
	  "(0 = A) ==> tyvars <RecType , 0> = A" tacfun;

val tyvars_rectype_rule
  = prove_goalw thy meta_defs
	  "[| tyvars t = A; tyvars <RecType, r> = A'; A Un A' = A''|] ==> \
	 \ tyvars <RecType , cons (<label, t>, r)> = A''" tacfun;

val tyvars_constype_empty_rule
  = prove_goalw thy meta_defs
	  "(0 = A) ==> tyvars <ConsType, 0, tyname> = A" tacfun;

val tyvars_constype_rule
  = prove_goalw thy meta_defs
 "[| tyvars t = A; tyvars <ConsType, s, tyname> = A'; A Un A' = A''|] ==> \
	\ tyvars <ConsType , <t,s>, tyname> = A''" tacfun;
 
val tyvars_typescheme_rule     
  = prove_goalw thy meta_defs
	  "[| tyvars t = B; B - A = C |] ==> tyvars <All, A, t> = C" tacfun;

val tyvars_typefcn_rule     
  = prove_goalw thy meta_defs
	  "[| tyvars t = B; Tyvars CE = D; B Un D - A = C |] ==> \
\          tyvars <<Lam, A, t>, CE> = C" tacfun;

val Tyvars_empty_rule     
  = prove_goalw thy meta_defs
	  "(0 = A) ==> Tyvars 0 = A" tacfun;

val Tyvars_varenv_conenv_rule     (* for variable and constructor envs *)
  = prove_goalw thy meta_defs
	  "[| tyvars t = A; Tyvars VE = A'; A Un A' = A''|] ==> \
\ Tyvars cons (<<Index, x>, t>, VE) = A''" tacfun;

val Tyvars_tyenv_rule     (* for type envs *)
  = prove_goalw thy meta_defs
 "[| tyvars typefcn = A; Tyvars  TE = A';  A Un A' = A'' |] ==> \
\  Tyvars cons (<<TyCon, tycon>, typefcn>, TE) = A''" tacfun;

val Tyvars_tuple_rule     (* for contexts and environments *)
  = prove_goalw thy meta_defs
	  "[| Tyvars c = A; Tyvars cs = A'; A Un A' = A'' |] ==>\
\          Tyvars <c, cs> = A''" tacfun;

val Tyvars_overwrite_init_rule 
   = prove_goalw thy [Tyvars_overwrite_init]
	  "(Tyvars VE) = A  ==>  Tyvars (VE0 @@ VE) = A" 
  (fn prems => [ares_tac prems 1]);

val close_rule     
  = prove_goalw thy meta_defs
	   "[| tyvars t = B; Tyvars A = B'; B - B' = B''; \
\              <All, B'', t> = C |] ==> \
\           close (A, <All, 0, t>) = C" tacfun;

val Close_empty_rule     
  = prove_goalw thy meta_defs
         "(0 = B) ==> Close (A, 0) = B" tacfun;

val Close_rule     
  = prove_goalw thy meta_defs
         "[| close (A, t) = ts; Close (A, VE) = B; \
\            (cons(<ide, ts>, B) = B' )|] ==> \
\       Close (A, cons(<ide, t>, VE)) = B'" tacfun;


val tyvars_pr_type_rule 
  = prove_goalw thy meta_defs
         "[| tyvars  t1 = A; tyvars t2 = B; A Un B = C |] ==> \
\       tyvars  (t1 [*] t2) = C" tacfun;

(* -------------------------------------------------------------------- *)
(*	 	Resolution Tactics	  				*)
(* -------------------------------------------------------------------- *)

val combine_rules = [env_combine_equality, context_combine_equality];

val All_rules =  [All, All_empty];

val Close_rules = [close_rule, Close_rule, Close_empty_rule];

val subst_rules = 
 [subst_rule, subst_seq_empty_rule, subst_seq_rule];

val As_rules = [As1_rule, As2_rule];

val ABS_rules = [ABS21_rule, ABS22_rule];

val combine_rules = [context_combine_equality, env_combine_equality,
  C0_combine_equality];

val tyvars_rules = 
 [tyvars_tyvar_rule, tyvars_funtype_rule, tyvars_rectype_empty_rule,
  tyvars_rectype_rule, tyvars_constype_empty_rule, tyvars_constype_rule,
  tyvars_pr_type_rule, tyvars_typescheme_rule, tyvars_typefcn_rule,
  tyvars_int, tyvars_bool]; 

val Tyvars_rules = 
 [Tyvars_empty_rule,     
  Tyvars_varenv_conenv_rule, 
  Tyvars_tyenv_rule, Tyvars_tuple_rule, Tyvars_overwrite_init_rule,
  Tyvars_T0, Tyvars_U0, 
  Tyvars_SE0, Tyvars_TE0, 
  Tyvars_VE0, Tyvars_EE0, Tyvars_C0]; 

val init_rules = 
   [SE0_type, TE0_type, VE0_type, EE0_type,
    apply_SE0, apply_TE0, apply_VE0, apply_EE0];

val Obj_rules = [Var_type, Con_type, ExCon_type] @ 
                 init_rules @ All_rules @ 
                 Close_rules @ subst_rules @ As_rules @ ABS_rules @
                 Tyvars_rules @ combine_rules @ VE0_mems @ TE0_mems @ set_rules;


(* -------------------------------------------------------------------- *)
(*	 	Solving Side Conditions.  				*)
(* -------------------------------------------------------------------- *)

val distinct_rules = 
  [var_con_distinct, 
   sym RS var_con_distinct,
   var_excon_distinct,
   sym RS var_excon_distinct,
   con_excon_distinct, 
   sym RS con_excon_distinct,
   UNIT_BOOL_distinct,
   sym RS UNIT_BOOL_distinct,
   UNIT_INT_distinct,
   sym RS UNIT_INT_distinct,
   UNIT_LIST_distinct,
   sym RS UNIT_LIST_distinct,
   BOOL_INT_distinct,
   sym RS BOOL_INT_distinct,
   BOOL_LIST_distinct,
   sym RS BOOL_LIST_distinct,
   INT_LIST_distinct,
   sym RS INT_LIST_distinct];

val Elab_cs = Kernel_cs addSEs distinct_rules;
 
(* -------------------------------------------------------------------- *)
(*	 Functions for Polymorphic Type Inference			*)
(* 	 These use the built-in (distinct) type variables a1, ..., a9	*)
(* -------------------------------------------------------------------- *)

local fun prep_a n = 
   Const("Pair", Type ("fun", [Type ("i", []), Type ("fun",
      [Type ("i", []), Type ("i", [])])])) 
   $ Const("_TyVar", Type ("i", [])) $ 
    (mk_tr ("a"^(string_of_int n)) "intyvar" []);
      fun list_from_to n m =
         if (n = m) then  [n]
            else
         n::(list_from_to (n+1) m);
in  fun newvars m n = 
  if m > n then [] 
  else  map prep_a (list_from_to m n)
end;

fun term_list_to_term [t] = t | 
    term_list_to_term (t::y) = t $ (term_list_to_term y);

(* -------------------------------------------------------------------- *)
(*	 set_tyvars:	takes a given theorem, and replaces each	*)
(*			scheme (as yet uninstantiated) variable 	*)
(*			occurring in a type variable by			*)
(*		        a corresponding type variable			*)
(* -------------------------------------------------------------------- *)

val marker = ref 1;

fun set_tyvars th =
  let  val prems = prems_of th;
       val outputs = 
             flat (map (fn (_ $ (_ $ _ $ x)) => 
                  (add_term_vars (x,[]))) prems);
       val p = if null prems then (concl_of th) else hd prems;
       val vars = filter
           (fn a => not (a mem outputs))
           (add_term_vars (p,[]));
       val cvars = map (Sign.cterm_of (sign_of thy)) vars;
       val cx    = map (Sign.cterm_of (sign_of thy)) 
                         (newvars (!marker) ((!marker) + length vars - 1))
   in
       (marker := (!marker) + length vars;
        instantiate ([], mk_pair_list cvars cx) th)
   end;

(* -------------------------------------------------------------------- *)
(*	 	Tactic for Solving Side Conditions			*)
(* -------------------------------------------------------------------- *)

 fun is con (Const (t, _)) = (con = t) |  
    is con (t $ t') = (is con t) |
    is con _ = false;

val obj_step_tac =  
 let val tac = (* DETERM o *) (kernel_step_tac
     (Obj_rules @ tyvars_rules @ Tyvars_rules) Elab_cs)  
 in
  SUBGOAL (fn (prem,i) => 
   (let val t = Logic.strip_assums_concl prem;
        val (_$ (_ $ (A $ B) $ _)) =  t 
    in 
      if (is "tyvars" A andalso is_schvar B) then 
         CHANGED (PRIMITIVE set_tyvars)
      else
          tac i
    end) 
    handle _ => tac i)
  end;

val obj_tac = DEPTH_SOLVE_1 o obj_step_tac;

end
end

