
(* -------------------------------------------------------------*)
(*		Semantic Objects for Elaboration		*)
(* ------------------------------------------------------------ *)

elab_obj = Kernel +

consts

    (* simple semantic objects  *)
    (* Note - TyVar already defined in Kernel/SML_Syntax.ml *)

   "_TyName"	::	 "i"	("TyName")    
   "_StrName"	::	 "i"	("StrName") 

    (* compund semantic objects  *)
   
   "_Type"	::	 "i"	("Type")    
   "_RecType"	::	 "i"	("RecType")    
   "_FunType"	::	 "i"	("FunType")    
   "_ConsType"	::	 "i"	("ConsType")    
   "_TypeFcn"	::	 "i"	("TypeFcn")    
   "_TypeScheme"::	 "i"	("TypeScheme")    
   "_All"	::	 "i"	("All")    
   "_Lam"	::	 "i"	("Lam")    
   "_Str"	::	 "i"	("Str")    
   "_StrEnv"	::	 "i"	("StrEnv")    
   "_TyStr"	::	 "i"	("TyStr")    
   "_TyEnv"	::	 "i"	("TyEnv")    
   "_ConEnv"	::	 "i"	("ConEnv")    
   "_VarEnv"	::	 "i"	("VarEnv")    
   "_ExConEnv"	::	 "i"	("ExConEnv")    
   "_Env"	::	 "i"	("Env")    
   "_TyNameSet"	::	 "i"	("TyNameSet")    
   "_TyVarSet"	::	 "i"	("TyVarSet")    
   "_Context"	::	 "i"	("Context")    

    (* initial context and components  *)

   "_C0"	::	 "i"	("C0")    
   "_T0"	::	 "i"	("T0")    
   "_U0"	::	 "i"	("U0")    
   "_SE0"	::	 "i"	("SE0")    
   "_TE0"	::	 "i"	("TE0")    
   "_VE0"	::	 "i"	("VE0")    
   "_EE0"	::	 "i"	("EE0")    

    (* semantic operations  *)

   env_combine		::	 "[i, i] => i"	("_ # _")    
   context_combine	::	 "[i, i] => i"	("_ ## _")    

    (* type schemes and type functions *)
    
   instance	::	 "[i, i] => o"	("_ [>] _")    
   tyvars	::	 "i=> i"	("tyvars _")    
   Tyvars	::	 "i=> i"	("Tyvars _")    
   Close	::	 "[i, i] => i"	("Close '(_ , _')")    
   close	::	 "[i, i] => i"	("close '(_ , _')")    
   As		::	 "[i, i] => o"	("_ As _")    
   ABS2		::	 "[i, i] => o"	 ("ABS '(_,_')")    
   ABS3		::	 "[i, i, i] => o"	 ("ABS '(_,_,_')")    
   subst	::	 "[i, i, i, i] => o"	("subst '(_,_,_,_')")    
   subst_seq	::	 "[i, i, i, i] => o"	("subst'_seq '(_, _, _, _')")
   alpha	::	 "i"	("alpha")    


    (* sequents for type inference *)

   Exp_Seq	::	 "[i,Exp,i] => o"	 ("(3_ /|- _ /~> _)")    
   ExpRow_Seq	::	 "[i,ExpRow,i] => o"	 ("(3_ /|- _ /~> _ [ExpRow])")
   Match_Seq	::	 "[i,Match,i] => o"	 ("(3_ /|- _ /~> _ [Match])")
   Pat_Seq	::	 "[i,Pat,i,i] => o"	 ("(3_ /||- _ /~> '(_,_') [Pat])")    
   PatRow_Seq	::	 "[i,PatRow,i,i] => o"	 ("(3_ /||- _ /~> '(_,_')  [PatRow])")    
   Ty_Seq	::	 "[i,Ty,i] => o"	 ("(3_ /|- _ /~> _    [Ty])")    
   TySeq_Seq	::	 "[i,TySeq,i] => o"	 ("(3_ /|- _ /~> _    [TySeq])")    
   TyRow_Seq	::	 "[i,TyRow,i] => o"	 ("(3_ /|- _ /~> _    [TyRow])")    
   Dec_Seq	::	 "[i,Dec,i] => o"	 ("(3_ /||- _ /~> _)")    
   ValBind_Seq	::	 "[i,ValBind,i] => o"	 ("(3_ /||- _ /~> _  [ValBind])")    
   TypBind_Seq	::	 "[i,TypBind,i] => o"	 ("(3_ /||- _ /~> _  [TypBind])")    
   DatBind_Seq	::	 "[i,DatBind,i,i] => o"	 ("(3_ /||- _ /~> _,_  [DatBind])")    
   ConBind_Seq	::	 "[i,i,ConBind,i] => o"	 ("(3_ , _ /||- _ /~> _  [ConBind])")    
   ExBind_Seq	::	 "[i,ExBind,i] => o"	 ("(3_ /||- _ /~> _  [ExBind])")    
   Program_Seq	::	 "[i,Program,i] => o"	 ("(3_ /||- /_ ~> _  [Program])")    

   (* predefined objects *)

   " int"	::	 "i"	("int")    
   " bool"	::	 "i"	("bool")    
   " exn"	::	 "i"	("exn")    
   " unit"	::	 "i"	("unit")    
   " int_name"	::	 "i"	("INT'_NAME")    
   " bool_name"	::	 "i"	("BOOL'_NAME")    
   " exn_name"	::	 "i"	("EXN'_NAME")    
   " unit_name"	::	 "i"	("UNIT'_NAME")    
   " list_name"	::	 "i"	("LIST'_NAME")    
   "pr_type"	::	 "[i, i] => i"	   ("_ [*] _")

(* -------------------------------------------------------------*)
(*		Definitions of Semantic Objects			*)
(*	       (see page 17 of the Definition of SML)		*)
(* ------------------------------------------------------------ *)

rules

    (*  Semantic Objects *)

Type_def 	  "Type == SUM x : {TyVar, RecType, FunType, ConsType}. x"
RecType_def	  "RecType == Lab -fin-> Type"
FunType_def	  "FunType == Type * Type"
ConsType_def	  "ConsType == list (Type) * TyName"
TypeFcn_def	  "TypeFcn  == {Lam} * list (TyVar) * Type"
TypeScheme_def	  "TypeScheme == {All} * Pow (TyVar) *  Type"
Str_def	 	  "Str == StrName * Env"
TyStr_def	  "TyStr == TypeFcn * ConEnv"
StrEnv_def	  "StrEnv == StrId -fin-> Str"
TyEnv_def	  "TyEnv == TyCon -fin-> TyStr"
ConEnv_def	  "ConEnv == Con -fin-> TypeScheme"
VarEnv_def	  "VarEnv == (SUM x : {Var, Con, ExCon}. x) -fin-> TypeScheme"
ExConEnv_def	  "ExConEnv == ExCon -fin-> Type"
Env_def	  	  "Env == StrEnv * TyEnv * VarEnv * ExConEnv"
TyNameSet_def	  "TyNameSet == Fin (TyName)"
TyVarSet_def	  "TyVarSet == Fin (TyVar)"
Context_def	  "Context == TyNameSet * TyVarSet * Env"

    (*  Operations on Environments *)

context_combine_def "<T,U,E> ## <T',U',E'> ==<T Un T', U Un U', E # E'>"
env_combine_def	    
  "<SE,TE,VE,EE> # <SE',TE',VE',EE'> == \
\    <SE @@ SE',TE @@ TE',VE @@ VE',EE @@ EE'>"

    (*  Predefined Quantities - defined by meta-level rewrites,  *)
    (*  rather than parse and print translations		 *)

int_def	 	  "int == <ConsType, 0, INT_NAME>"
bool_def	  "bool == <ConsType, 0, BOOL_NAME>"
exn_def	 	  "exn == <ConsType, 0, EXN_NAME>"
unit_def	  "unit == <ConsType, 0, UNIT_NAME>"
pr_type_def	  "t1 [*] t2 == <RecType, {<n1,t1>,<n2,t2>}>"
int_tyname	  "INT_NAME : TyName"
bool_tyname	  "BOOL_NAME : TyName"
exn_tyname	  "EXN_NAME : TyName"
unit_tyname	  "UNIT_NAME : TyName"
list_tyname	  "LIST_NAME : TyName"

    (*  Rules expressing the fact that identifiers are distinct sets  *)

var_con_distinct	  "<Var, x> = <Con, y> ==> P"
var_excon_distinct	  "<Var, x> = <ExCon, y> ==> P"
con_excon_distinct	  "<Con, x> = <ExCon, y> ==> P"
UNIT_BOOL_distinct	  "UNIT = BOOL ==> P"
UNIT_INT_distinct	  "UNIT = INT ==> P"
UNIT_LIST_distinct	  "UNIT = LIST ==> P"
BOOL_INT_distinct	  "BOOL = INT ==> P"
BOOL_LIST_distinct	  "BOOL = LIST ==> P"
INT_LIST_distinct	  "INT = LIST ==> P"

    (*  Initial Context  *)
C0_def        "C0 == <T0, U0, SE0, TE0, VE0, EE0>"
T0_def        "T0 == 0"
U0_def        "U0 == 0"
SE0_def       "SE0 == 0"
TE0_def
   "TE0 ==   \
\  {  <<TyCon, UNIT>, <Lam, 0, RecType, 0>, 0>,                  \
\     <<TyCon, BOOL>, <Lam, 0, bool>,                            \
\       { <<Con, true>,  bool>, <<Con, false>, bool> }>,         \
\     <<TyCon, INT>, <Lam, 0, int>, 0>, 			 \
\     <<TyCon, List>,                                            \
\       <Lam, {alpha}, ConsType, <<TyVar,alpha>, 0>, LIST_NAME>, \
\       {<<Con, nil>,   <All, {alpha},                           \
\           <ConsType, <<TyVar,alpha>, 0>, List>>>,              \
\        <<Con, CONS>,                                           \
\         <All, {alpha}, <FunType,                               \
\      <TyVar, alpha> 				                 \
\	[*] <ConsType, <<TyVar,alpha>, 0>, List>,	         \
\      <ConsType, <<TyVar,alpha>, 0>, List>>>>}>,                \
\     <<TyCon, EXN>, <Lam, 0, exn>, 0> }"

VE0_def        
  "VE0 ==   {  <<Var, plus>,   		          	  \
\             <All, 0, <FunType, int [*] int, int>>>,     \
\     <<Var, minus>,   				          \
\             <All, 0, <FunType, int [*] int, int>>>,     \
\     <<Var, times>,  				          \
\             <All, 0, <FunType, int [*] int, int>>>,     \
\     <<Var, zero>,   <All, 0, <FunType, int, bool>>>,    \
\     <<Var, not>,    <All, 0, <FunType, bool, bool>>>,   \
\     <<Var, head>,                                       \
\      <All, {alpha}, <FunType,                           \
\          <ConsType, <<TyVar,alpha>, 0>, List>,          \
\                      <TyVar, alpha>>>>,                 \
\     <<Var, tail>,                                       \
\         <All, {alpha}, <FunType,                        \
\          <ConsType, <<TyVar,alpha>, 0>, List>,          \
\          <ConsType, <<TyVar,alpha>, 0>, List>>>>,       \
\     <<Var, null>,                                       \
\         <All, {alpha}, <FunType,                        \
\          <ConsType, <<TyVar,alpha>, 0>, List>, bool>>>, \
\     <<Con, true>,    <All, 0, bool>>,                   \
\     <<Con, false>,   <All, 0, bool>>,                   \
\     <<Con, nil>,   <All, {alpha},                       \
\           <ConsType, <<TyVar,alpha>, 0>, List>>>,       \
\     <<Con, CONS>,                                       \
\         <All, {alpha}, <FunType,                        \
\      <TyVar, alpha> 				          \
\	[*] <ConsType, <<TyVar,alpha>, 0>, List>,	  \
\      <ConsType, <<TyVar,alpha>, 0>, List>>>> }"
EE0_def        "EE0 == 0"
C0_type        "C0 : Context"

    (*  Various Equalities - for Object Level Rewriting *)

    (*  Type Schemes and Instances *)

All_empty  "<All, 0, t> [>] t"
All	   "<All, A, t (t')> [>] u  ==> \
\                    <All, cons(a,A),  t(<TyVar, a>)> [>] u"


    (*  Rules for finding the set of free type variables in a given type *)

tyvars_tyvar	   	  "tyvars <TyVar, a> == {a}"  
tyvars_funtype	 	  "tyvars <FunType , t, t'> == tyvars t Un tyvars t'"
tyvars_rectype_empty	  "tyvars <RecType , 0> == 0" 
tyvars_rectype	  	  "tyvars <RecType , cons (<label, t>, r)> \
\    				== tyvars t Un tyvars <RecType, r>"
tyvars_constype_empty	  "tyvars <ConsType, 0, tyname> == 0"
tyvars_constype	  	  "tyvars <ConsType , <t,s>, tyname> == \
\ 				tyvars t Un tyvars <ConsType, s, tyname>"
 
   (*  Rules for type schemes and type functions *)

tyvars_typescheme	  "tyvars <All, A, t> == tyvars t - A"  
tyvars_typefcn	          "tyvars <<Lam, A, t>, CE> == \
\                              tyvars t Un Tyvars CE - A"  

   (*  Rules for environments, contexts *)

Tyvars_empty  		  "Tyvars 0 == 0"   
                               (* for variable and constructor envs *)
Tyvars_varenv_conenv	  "Tyvars cons (<<Index, x>, t>, VE) == \
\				tyvars t Un Tyvars VE"  
(*
Tyvars_tyenv	          "Tyvars cons (<<TyCon, tycon>, theta, CE>, TE) == \
\		                tyvars theta Un Tyvars CE Un Tyvars TE"  
*)
Tyvars_tyenv	          "Tyvars cons (<<TyCon, tycon>, typefcn>, TE) == \
\		                tyvars typefcn Un Tyvars TE"  
                                  (* for type envs *)
Tyvars_tuple	          "Tyvars <c, cs> == Tyvars c Un Tyvars cs" 
                                  (* for contexts and environments *)

(* we should be able to prove this result : *)

Tyvars_overwrite_init     "Tyvars (VE0 @@ VE) == Tyvars VE"

   (*  Closures *)

close	            	 "close (A, <All, 0, t>) == \
\				<All, tyvars t - Tyvars A, t>"   
Close_empty              "Close (A, 0) == 0"   
Close              "Close (A, cons(<ide, t>, VE)) == \
\			cons (<ide, close (A, t)>, Close (A, VE))"


subst_rule	   	  "subst (t(<TyVar,a>), t', a, t(t'))"
subst_seq_empty_rule	  "subst_seq (t, 0, 0, t)"
subst_seq_rule            
 "[| subst (t, t', a, t''); subst_seq (t'',typeseq, tyvarseq, t''') |] ==> subst_seq (t, <t', typeseq>, cons (a, tyvarseq), t''')"

As1_rule           "0 As 0"
As2_rule           "tyvarseq As typeseq ==>  <a,tyvarseq> As <<TyVar, a>,  typeseq>"

ABS21_rule         "ABS (0,0)"
ABS22_rule         
 "tyvarseq As typeseq ==> ABS ({<<TyCon, tycon>, <Lam, tyvarseq, ConsType, typeseq, tyname>, CE>},   {<<TyCon, tycon>, <Lam, tyvarseq, ConsType, typeseq, tyname>, 0>})"


end

 

