(* -------------------------------------------------------------------- *)
(*									*)
(*	Eval.ML:							*)
(*	Theorem prover for Evaluation					*)
(*					 				*)
(* -------------------------------------------------------------------- *)

structure EvalProver =
struct
local open eval_obj eval_obj_lemmas eval_program in 

val intr_rules_data =
  [ 
    ("Scon", [["v'"]]),
    ("Var_", [["v'"]]),
    ("Con", [["v'"]]),
    ("ExCon", [["v'"]]),
    ("Empty_Record", [["v'"]]),
    ("Record", [["v'"]]),
    ("Let", [[]]),
    ("Bracket", [[]]),
    ("ExpRow1", [["record'"]]),
    ("ExpRow2", [["record'"]]),
    ("Apply", [[]]),
(*  ("Handle", [["v'"]]), *)
(*  ("Raise", [["v'"]]),  *)
    ("Fn", [[]]),
    ("Match_", [["v'"], ["v'"]]),
    ("Mrule", [["v'"], ["v'"]]),
    ("Val", [["E'"]]),
    ("Exception", [["E'"]]),
    ("Local", [[]]),
    ("Nothing", [["E'"]]),
    ("Comp1", [[]]),
    ("Comp2", [[]]),
    ("ValBind1", [[]]),
    ("ValBind2", [[]]),
    ("ValBind3", [["pack"]]),
    ("ValBind4", [["pack"]]),
    ("Rec", [[]]),
    ("Wild_Pat", [["VE'"]]),
    ("Scon_Pat", [["VE'"],["VE'"]]),
    ("Var_Pat", [["VE'"]]),
    ("Con_Pat", [["VE'"], ["VE'"]]),
    ("ExCon_Pat", [["VE'"], ["VE'"]]),
    ("Empty_Record_Pat", [["VE'"]]),
    ("Record_Pat", [[]]),
    ("Bracket_Pat", [[]]),
    ("Wild_PatRow", [["VE'"]]),
    ("PatRow1", [[]]),
    ("PatRow2", [["VE'"], [], ["VE'"]]),
    ("Con_Val_Pat", [["VE'"], ["VE'"]]),
    ("ExCon_Val_Pat", [["VE'"], ["VE'"]]),
    ("Program1", [[]]),
    ("Program2", [[]])
];

val unflattened_intr_rules = 
     map (get_intr_rules thy)
         intr_rules_data;

val intr_rules = 
 flat unflattened_intr_rules;

(* -------------------------------------------------------------------- *)
(*	Some Derived Rules for the Initial Context			*)
(* -------------------------------------------------------------------- *)

fun tacfun prems = 
	     [cut_facts_tac prems 1,
              resolve_tac intr_rules 1,
              REPEAT (ares_tac ([apply_VE0, refl]) 1)];

val Var_init = prove_goalw thy [E0_def]
 "<<Var,x>, v> : VE0 ==> E0 |- var x ~> <Val, v>" tacfun;

(* -------------------------------------------------------------------- *)
(*	step tactics:	applies a language rule once, and safely	*)
(* -------------------------------------------------------------------- *)

val eval_step_tac =  
 SUBGOAL (fn (prem,i) => 
  ares_tac ([Var_init] @ VE0_mems) i ORELSE
  let val (_$ (_ $ phrase $ _)) = Logic.strip_assums_concl prem in
   if (is_schvar phrase) then no_tac (* indeterminate goal *) 
   else resolve_tac intr_rules i
  end
  handle _ => no_tac);

(* -------------------------------------------------------------------- *)
(*	 eval_tac:	general-purpose tactic for proving goals by	*)
(*			depth-first search				*)
(*	Note that resolution with apply_intr_rules must be done		*)
(*	within the highest-level depth-first search (not in obj_tac)	*)
(* -------------------------------------------------------------------- *)

val step_tac = eval_step_tac ORELSE' apply_tac ORELSE' obj_tac;

val evaluate_tac  = REPEAT1 (step_tac 1);

val eval_tac  = DEPTH_SOLVE evaluate_tac;

(* -------------------------------------------------------------------- *)
(*		 Interface Functions					*)
(* -------------------------------------------------------------------- *)

fun exp_eval phrase 
 = (prs "\n\n"; goal thy ("E0 |- "^phrase^" ~> ?v"));

fun dec_eval phrase 
 = (prs "\n\n"; goal thy ("E0 ||- "^phrase^" ~> ?E"));

fun program_eval phrase
 = (prs "\n\n"; goal thy ("E0 ||- "^phrase^" ~> ?E [Program]"));

(* -------------------------------------------------------------------- *)
(*	 Pretty printing flags 						*)
(* -------------------------------------------------------------------- *)

val show_envs = ref false;
val show_full_values = ref false;

fun Envtr' [SE,VE,EE] =  
  if (!show_envs) then raise Match else Const (" Env", dummyT); 

fun tr' [x] = 
  if (!show_full_values) then raise Match else x;


val thy = thy;

end
end;
