(* -----------------------------------------------------------	*)
(*		Testing the Language Syntax	 		*)
(* -----------------------------------------------------------	*)

open SML_Syntax;
val echo = prin o read;

goal thy "fred";

print_depth 30;

echo "[]";
echo "var x ++ var y";
echo "var x ::[]";
echo "[var x]";
echo "var x::[var y]";
echo "var x::[var y, var z]";
echo "var x::var L";
echo "(- var x, var y -)";
echo "val (- var x, var y -) = var z";
echo "var x : int <*> int";
echo "if var x then var y else var z";
echo "case var x of var y => var z";
echo "var x orelse var y"; 

echo "var x andalso var y";
echo "var x;";
echo "int";
echo "bool";
echo "unit";
echo "exn";



(* -----------------------------------------------------------	*)
(*		Testing function application	 		*)
(* -----------------------------------------------------------	*)

open Kernel;

goal thy 
"{<x1,a>, <x2,b>, <x3,c>, <x4, d>} `x1 = ?t";
by (kernel_tac [] 1 );

goal thy
 "{<x1,a>, <x2,b>, <x3,c>, <x4, d>} `x2 = ?t";
by (kernel_tac [] 1);

goal thy 
"{<x1,a>, <x2,b>, <x3,c>, <x4, d>} `x3 = ?t";
by (kernel_tac [] 1);

goal thy 
"{<x1,a>, <x2,b>, <x3,c>, <x4, d>} `x4 = ?t";
by (kernel_tac [] 1);

(* -----------------------------------------------------------	*)
(*		Testing overwrite/apply simplification 		*)
(* -----------------------------------------------------------	*)

goal thy 
"{<x1,a>, <x2,b>, <x3,c>} @@ {<x2,d>, <x4, e>} ` x1 = ?t";  (* expect a *)
by (kernel_tac [] 1);

goal thy 
 "{<x1,a>, <x2,b>, <x3,c>} @@ {<x2,d>, <x4, e>} ` x2 = ?t";  (* expect d *)
by (kernel_tac [] 1);

goal thy 
 "{<x1,a>, <x2,b>, <x3,c>} @@ {<x2,d>, <x4, e>} ` x3 = ?t";  (* expect c *)
by (kernel_tac [] 1);

val prems = goal thy 
 "{<x1,a>, <x2,b>, <x3,c>} @@ {<x2,d>, <x4, e>} ` x4 = ?t";  (* expect e *)
by (kernel_tac [] 1);

goal thy 
 "{<x1,a>, <x2,b>, <x3,c>} @@ {<x2,d>, <x4, e>} @@ \
\ {<x5, f>}` x1 = ?t";  

by (kernel_tac [] 1);

goal thy 
 "{<x1,a>, <x2,b>, <x3,c>} @@ {<x2,d>, <x4, e>} @@ \
\ {<x5, f>} @@ {<x6,g>}` x5 = ?t";  

by (kernel_tac [] 1);

(* -----------------------------------------------------------	*)
(*		Testing overwrite simplification 		*)
(* -----------------------------------------------------------	*)


goal thy  "{<x1,a>, <x2, b>} @@  {<x3, e>}  = ?t";

by (kernel_tac [] 1);

goal thy  "{<x1,a>, <x2, b>} @@  {<x2,d>, <x3, e>}  = ?t"; 

by (kernel_tac [] 1);

goal thy " {<x1,a>, <x2, b>, <x3, h>} @@ {<x2,d>, <x4, e>}  = ?t"; 

by (kernel_tac [] 1);

goal thy " {<x1,a>, <x2, b>, <x3, h>} @@ {<x2,d>, <x3, e>}  = ?t"; 

by (kernel_tac [] 1);
