(* -------------------------------------------------------------*)
(*		Identifiers.ML					*)
(*		- Modelling the SML Identifiers in SML		*)
(* ------------------------------------------------------------ *)


identifiers = Tags +

consts

  (* identifiers are sets *)

   "_Var"	::	 "i"	("Var")     (* value variables       *)
   "_Con"	::	 "i" 	("Con")     (* value constructors    *)
   "_ExCon" 	::	 "i"	("ExCon")   (* exception variables   *)
   "_TyVar"	::	 "i"	("TyVar")   (* type variables        *)
   "_TyCon"	::	 "i"	("TyCon")   (* type constructors     *)
   "_Lab"	::	 "i"	("Lab")     (* record labels         *)
   "_StrId"	::	 "i"	("StrId")   (* structure identifiers *)

  (* Built-in Variables *)

   " plus"	::	 "i"            ("plus")   
   " minus"	::	 "i"            ("minus")   
   " times"	::	 "i"            ("times")   
   " less"	::	 "i"            ("less")   
   " zero"	::	 "i"            ("zero")   
   " not"	::	 "i"            ("not")   
   " head"	::	 "i"            ("head")   
   " tail"	::	 "i"            ("tail")   
   " null"	::	 "i"            ("null")   

  (* Built-in Value Constructors *)

   " true"	::	 "i"            ("true")   
   " false"	::	 "i"            ("false")   
   " nil"	::	 "i"            ("nil")   
   " CONS"	::	 "i"            ("CONS")   

 (* Predefined Type Constructors *)

   "int_con"	::	 "i"            ("INT")  
   "bool_con"	::	 "i"            ("BOOL")  
   "unit_con"	::	 "i"            ("UNIT")  
   "exn_con"	::	 "i"            ("EXN")  
   "list_con"	::	 "i"            ("List")  

 (* Predefined Exception Constructors *)

   "ERROR_excon"	::	 "i"     ("ERROR")  
   "MATCH_excon"	::	 "i"     ("MATCH")  
   "BIND_excon"		::	 "i"     ("BIND")  

 (* The variables x1, ..., x9 will be fixed and distinct *)

   " x1"	::	 "i"            ("x1")   
   " x2"	::	 "i"            ("x2")   
   " x3"	::	 "i"            ("x3")   
   " x4"	::	 "i"            ("x4")   
   " x5"	::	 "i"            ("x5")   
   " x6"	::	 "i"            ("x6")   
   " x7"	::	 "i"            ("x7")   
   " x8"	::	 "i"            ("x8")   
   " x9"	::	 "i"            ("x9")   

 (* the variables f1, ..., f9 can be used for functions *)

   " f1"	::	 "i"            ("f1")   
   " f2"	::	 "i"            ("f2")   
   " f3"	::	 "i"            ("f3")   
   " f4"	::	 "i"            ("f4")   
   " f5"	::	 "i"            ("f5")   
   " f6"	::	 "i"            ("f6")   
   " f7"	::	 "i"            ("f7")   
   " f8"	::	 "i"            ("f8")   
   " f9"	::	 "i"            ("f9")   

 (* the variable "it" is used to remember previous expression evaluations *)

   " it"	::	 "i"            ("it")   

 (* a1 ... a9 are type variables *)

   " a1"	::	 "i"            ("a1")   
   " a2"	::	 "i"            ("a2")   
   " a3"	::	 "i"            ("a3")   
   " a4"	::	 "i"            ("a4")   
   " a5"	::	 "i"            ("a5")   
   " a6"	::	 "i"            ("a6")   
   " a7"	::	 "i"            ("a7")   
   " a8"	::	 "i"            ("a8")   
   " a9"	::	 "i"            ("a9")   

 (* T1 ... T9 are type constructors *)

   " T1"	::	 "i"            ("T1")   
   " T2"	::	 "i"            ("T2")   
   " T3"	::	 "i"            ("T3")   
   " T4"	::	 "i"            ("T4")   
   " T5"	::	 "i"            ("T5")   
   " T6"	::	 "i"            ("T6")   
   " T7"	::	 "i"            ("T7")   
   " T8"	::	 "i"            ("T8")   
   " T9"	::	 "i"            ("T9")   

 (* S1 ... S9 are structure identifiers *)

   " S1"	::	 "i"            ("S1")   
   " S2"	::	 "i"            ("S2")   
   " S3"	::	 "i"            ("S3")   
   " S4"	::	 "i"            ("S4")   
   " S5"	::	 "i"            ("S5")   
   " S6"	::	 "i"            ("S6")   
   " S7"	::	 "i"            ("S7")   
   " S8"	::	 "i"            ("S8")   
   " S9"	::	 "i"            ("S9")   

 (* The distinct labels n1 and n2 are reserved for pairs *)

   " n1"	::	 "i"            ("n1")   
   " n2"	::	 "i"            ("n2")   

 (* ... while p1, ..., p6 can be used for general labels *)

   " p1"	::	 "i"            ("p1")   
   " p2"	::	 "i"            ("p2")   
   " p3"	::	 "i"            ("p3")   
   " p4"	::	 "i"            ("p4")   
   " p5"	::	 "i"            ("p5")   
   " p6"	::	 "i"            ("p6")   
   " p7"	::	 "i"            ("p7")   

 (* ... and c1, ..., c9 can be used for general value constructors *)

   " c1"	::	 "i"            ("c1")   
   " c2"	::	 "i"            ("c2")   
   " c3"	::	 "i"            ("c3")   
   " c4"	::	 "i"            ("c4")   
   " c5"	::	 "i"            ("c5")   
   " c6"	::	 "i"            ("c6")   
   " c7"	::	 "i"            ("c7")   
   " c8"	::	 "i"            ("c8")   
   " c9"	::	 "i"            ("c9")   

 (* Injections *)

   "invar"	::	 "i => i"       ("invar _")  
   "intyvar"	::	 "i => i"       ("intyvar _")  
   "intycon"	::	 "i => i"       ("intycon _")  
   "instrid"	::	 "i => i"       ("instrid _")  
   "inlabel"	::	 "i => i"       ("inlabel _")  
   "incon"	::	 "i => i"       ("incon _")  
   "inexcon"	::	 "i => i"       ("inexcon _")  


rules

   invar_inject		"invar m = invar n ==> m = n"
   invar_type		"invar m : Var"
   intyvar_inject	"intyvar m = intyvar n ==> m = n"
   intyvar_type		"intyvar m : TyVar"
   intycon_inject	"intycon m = intycon n ==> m = n"
   intycon_type		"intycon m : TyCon"
   instrid_inject	"instrid m = instrid n ==> m = n"
   instrid_type		"instrid m : StrId"
   inlabel_inject	"inlabel m = inlabel n ==> m = n"
   inlabel_type		"inlabel m : Lab"
   incon_inject		"incon m = incon n ==> m = n"
   incon_type		"incon m : Con"
   inexcon_inject	"inexcon m = inexcon n ==> m = n"
   inexcon_type		"inexcon m : ExCon"

end

ML


(* -------------------------------------------------------------------- *)
(*			 Parse  Translations  				*)
(* -------------------------------------------------------------------- *)

(* -------------------------------------------------------------------- *)
(*   	  Predefined Variables, Labels and Built-Ins 			*)
(* -------------------------------------------------------------------- *)

(* xn (c, n) should only be given an n > 0. WE restrict to n > 1,
as we don't want 0 to be used as a constant *)

fun mk_xn injection n = 
let fun xn (c, 1) = c
      | xn (c, n) = 
          if n mod 2 = 0 then xn 
              (Const ("black", 
                Type ("fun", [Type ("i", []), Type ("i", [])])) $ c, n div 2)
                         else xn 
              (Const ("white", 
                Type ("fun", [Type ("i", []), Type ("i", [])])) $ c, n div 2) 
in Const(injection, 
         Type ("fun", [Type ("i", []), Type ("i", [])])) $ 
    xn(Const("0", Type ("i", [])), n+2)
end;

fun pos x (y::ys) = if x = y then 0 else 1 + pos x ys;

fun mk_ide_list "invar"   = 
      ["x1", "x2", "x3", "x4", "x5", "x6", "x7", "x8", "x9",
       "f1", "f2", "f3", "f4", "f5", "f6", "f7", "f8", "f9", "it",
       "plus", "minus", "times", 
       "less", "zero", "not", 
       "head", "tail", "null"] |
    mk_ide_list "intyvar"=
      ["a1", "a2", "a3", "a4", "a5", "a6", "a7", "a8", "a9"] |
    mk_ide_list "intycon"=
      ["T1", "T2", "T3", "T4", "T5", "T6", "T7", "T8", "T9"] |
    mk_ide_list "instrid"=
      ["S1", "S2", "S3", "S4", "S5", "S6", "S7", "S8", "S9"] |
    mk_ide_list "inlabel"=
      ["n1", "n2",  (* reserved for pair translations *)
       "p1", "p2", "p3", "p4", "p5", "p6", "p7"] |
    mk_ide_list "incon" =
      ["true", "false", 
       "nil", "CONS", "c1", "c2", "c3", "c4", "c5", "c6",
       "c7", "c8", "c9"] |
    mk_ide_list "inexcon" =
      ["ERROR", "MATCH", "BIND"];

local 
fun mk_xn_list (injection, 0, n) = []
  | mk_xn_list (injection, m, n) = (mk_xn injection n)
                                 ::(mk_xn_list (injection, m-1, n+1))
in
val 
 invar_consts = mk_xn_list ("invar", length (mk_ide_list "invar"), 0)
and 
 intyvar_consts = mk_xn_list ("intyvar", length (mk_ide_list "intyvar"), 0)
and 
 intycon_consts = mk_xn_list ("intycon", length (mk_ide_list "intycon"), 0)
and 
 instrid_consts = mk_xn_list ("instrid", length (mk_ide_list "instrid"), 0)
and 
 inlabel_consts = mk_xn_list ("inlabel", length (mk_ide_list "inlabel"), 0)
and 
 incon_consts = mk_xn_list ("incon", length (mk_ide_list "incon"), 0)
and 
 inexcon_consts = mk_xn_list ("inexcon", length (mk_ide_list "inexcon"), 0)
end;

fun mk_tr name injection = 
  fn [] => mk_xn injection (pos name (mk_ide_list injection));

val parse_translation = 
let    fun mk_pr_list [] [] = [] | 
        mk_pr_list (x::xs) (y::ys) = (x,y)::(mk_pr_list xs ys)
       fun mk_name x = " "^x
       fun mk_tr' x = (fn [] => x)
in
  (mk_pr_list (map mk_name (mk_ide_list "invar"))
              (map mk_tr' invar_consts))
@
  (mk_pr_list (map mk_name (mk_ide_list "intyvar"))
              (map mk_tr' intyvar_consts))
@
  (mk_pr_list (map mk_name (mk_ide_list "intycon"))
              (map mk_tr' intycon_consts))
@
  (mk_pr_list (map mk_name (mk_ide_list "instrid"))
              (map mk_tr' instrid_consts))
@
  (mk_pr_list (map mk_name (mk_ide_list "inlabel"))
              (map mk_tr' inlabel_consts))
@
  (mk_pr_list (map mk_name (mk_ide_list "incon"))
              (map mk_tr' incon_consts))
@
  (mk_pr_list (map mk_name (mk_ide_list "inexcon"))
              (map mk_tr' inexcon_consts))
end;


(* -------------------------------------------------------------------- *)
(*			 Print  Translations  				*)
(* -------------------------------------------------------------------- *)

(* -------------------------------------------------------------------- *)
(*   	  Predefined Variables, Labels and Built-Ins 			*)
(* -------------------------------------------------------------------- *)

fun get_n tm =
let fun get_bin_n (Const ("black",_) $ t, n) = get_bin_n (t, 2*n)
  |     get_bin_n (Const ("white",_) $ t, n) = get_bin_n (t, 2*n + 1)
  |     get_bin_n (_, n) = n
in
get_bin_n (tm, 1) - 2 (* Compensate the +2 in mk_xn *)
end;


(* needs to be defined for PolyML - with 0 numbering first elt of list *)

fun nth (x::xs, 0) = x |
    nth (x::xs, n) = nth (xs, n-1);

fun mk_tr' injection [t] = 
  Const (" "^(nth (mk_ide_list injection, get_n t)), dummyT)
                      handle _ => t;

val print_translation =      [("invar", mk_tr' "invar"),
			      ("intyvar", mk_tr' "intyvar"),
			      ("intycon", mk_tr' "intycon"),
			      ("instrid", mk_tr' "instrid"),
			      ("inlabel", mk_tr' "inlabel"),
			      ("incon", mk_tr' "incon"),
			      ("inexcon", mk_tr' "inexcon")];


