
(* -------------------------------------------------------------*)
(*		Overwrite.ML: 					*)
(* ------------------------------------------------------------ *)

(* -------------------------------------------------------------------- *)
(*	Derived Rules for overwrite 					*)
(* -------------------------------------------------------------------- *)

open Overwrite;


val gen_fun_disjoint_Un = prove_goal thy 
"[| f : A -> B; g : C -> D; A Int C = 0 ; \
\    E = A Un C; F = B Un D |] ==> f Un g : E -> F" 
(fn prems => [asm_simp_tac (ZF_ss addsimps prems @ [fun_disjoint_Un]) 1]);

val shift_tac = DETERM o (etac revcut_rl);
val disch_tac = DETERM o (etac rev_mp);

val overwriteI1 = prove_goalw thy [overwrite_def] 
 "<x, y>:g ==> <x, y>: f @@ g"
(fn prems => [REPEAT (ares_tac (UnI2 :: prems) 1)]);

val overwriteI2 = prove_goalw thy [overwrite_def] 
  "[| <x, y> : f; ~x : domain(g) |] ==> <x, y>: f @@ g"
(fn prems => 
 [REPEAT (ares_tac (prems@ [CollectI, UnI1]) 1 ORELSE 
   simp_tac ZF_ss 1)]);

val not_overwriteI = prove_goalw thy [overwrite_def]
  "[| ~<x, y> : g; ~<x, y> : f |] ==> ~<x, y> : f @@ g"
(fn prems => [cut_facts_tac prems 1,
              fast_tac ZF_cs 1]);

val not_consI = prove_goal thy "[| ~x=h; ~x:t |] ==> ~x: cons(h, t)"
(fn prems => [cut_facts_tac prems 1,
              fast_tac ZF_cs 1]);

val not_cons_domain_I = prove_goal thy 
 "[| ~ x = y; ~ x : domain(f) |] ==> ~x : domain(cons(<y,a>, f))"
(fn prems => [cut_facts_tac prems 1,
              fast_tac ZF_cs 1]);

val functional_subset =   (* this proof is very delicate *)
prove_goal thy "[|f:A -> B; g <= f |] ==>  g: domain(g) -> range(g)"
(fn [p1, p2] => 
  [cut_facts_tac [p1, p2] 1, 
   rtac range_of_fun 1,
   rtac PiI 1, etac PiE 1, etac domainE 2, rtac ex1I 2, atac 2,
   fast_tac ZF_cs 1, flexflex_tac, dtac subsetD 1, atac 1, 
   cut_facts_tac [p2] 1, dtac subsetD 1,
   shift_tac 1, shift_tac 1, atac 1, rtac apply_equality2 1, atac 3, atac 1, 
   atac 1]);

val functional_Collect = prove_goal thy 
 "f:A -> B ==> {x:f. P(x)}: domain({x:f. P(x)}) -> range({x:f. P(x)})"
(fn prems => 
 [REPEAT  (ares_tac (prems @ [functional_subset, subsetI]) 1 
       ORELSE dtac CollectD1 1)]);



val [p1, p2]  = 
goalw   thy [overwrite_def]
   "[| f : A -> B; g : C -> D |] ==> f @@ g : domain(f @@ g) -> range(f @@ g)";
by (cut_facts_tac [p1, p2 RS range_of_fun] 1);
br gen_fun_disjoint_Un 1;
br functional_Collect 1;
ba 1;
ba 1;
by (asm_simp_tac (ZF_ss addsimps [domain_of_fun]) 1);
br equalityI 1;
br subsetI 1;
be IntE 1;
be domainE 1;
be CollectE 1;
be notE 1;
by (asm_simp_tac (ZF_ss) 1);
br empty_subsetI 1; (*14*)

br equalityI 1;
br subsetI 1;
be domainE 1;
be UnE 1;
br UnI1 1;
br domainI 1;
ba 1;
br UnI2 1;
br domain_type 1; ba 1; ba 1; (* 25*)

br subsetI 1;
be UnE 1;
be domainE 1;
br domainI 1;
br UnI1 1;
ba 1;
br domainI 1;
br UnI2 1;
br apply_Pair 1;
ba 1;
ba 1; (* 36 *)

br equalityI 1;
br subsetI 1;
be rangeE 1;
be UnE 1;
br UnI1 1;
br rangeI 1;
ba 1;
br UnI2 1;
by (res_inst_tac [("a", "xa"), ("b", "x")] range_type 1);
ba 1; ba 1;

br subsetI 1;
be UnE 1;
be rangeE 1;
br rangeI 1;
br UnI1 1;
ba 1;
be rangeE 1;
br rangeI 1;
br UnI2 1;
ba 1;
val overwrite_type = range_of_fun RS (result());

val overwrite0L = prove_goalw thy [overwrite_def] "0 @@ f = f"
(fn [] =>  [fast_tac eq_cs 1]);

val overwrite0R = prove_goalw thy [overwrite_def] "f @@ 0 = f"
(fn [] =>  [fast_tac eq_cs 1]);


val cases = excluded_middle RS disjE;

fun cases_tac tm = res_inst_tac [("Q", tm)] cases 1;

goalw thy [overwrite_def] "domain (f @@ g) = domain(f) Un domain(g)";
br equalityI 1;

br subsetI 1;
be domainE 1;
be UnE 1;
br UnI1 1;
be CollectE 1;
br domainI 1;
ba 1;
br UnI2 1;
br domainI 1;
ba 1;  (* 11 *)

br subsetI 1;;
be UnE 1;
be domainE 1;
by (cases_tac "fst(<x,y>) : domain(g)");
br domainI 1;
br UnI1 1;
br CollectI 1;
ba 1;  
ba 1; (* 20 *)

by (shift_tac 1);
by (disch_tac 1);
by (simp_tac ZF_ss 1);
br impI 1;
be domainE 1;
br domainI 1;
br UnI2 1;
ba 1;  (* 28 *)


be domainE 1;
br domainI 1;
br UnI2 1;
ba 1;  

val domain_overwrite = result();

val not_domainI = prove_goal thy
  "[| ~x : domain (f); ~x : domain(g) |] ==> ~x : domain (f @@ g)"
(fn prems => [simp_tac (ZF_ss addsimps (prems @ [domain_overwrite])) 1]);
 
goalw thy [overwrite_def] "(f @@ g) @@ h = f @@ (g @@ h)";
br equalityI 1;
br subsetI 1;
be UnE 1;
be CollectE 1;
be UnE 1;
be CollectE 1;
br UnI1 1;
br CollectI 1;
ba 1;
by (fold_tac [overwrite_def]);
br not_domainI 1;
ba 1;
ba 1;
bw overwrite_def;
br UnI2 1;
br UnI1 1;
br CollectI 1;
ba 1;
ba 1;
br UnI2 1;
br UnI2 1;
ba 1; 	(* 21 *)

br subsetI 1;
be UnE 1;
be CollectE 1;
br UnI1 1;
br CollectI 1;
br UnI1 1;
br CollectI 1;
ba 1;
by (fold_tac [overwrite_def]);
be swap 1;
by (simp_tac (ZF_ss addsimps [domain_overwrite]) 1);
by (fast_tac FOL_cs 1);
be swap 1;
by (simp_tac (ZF_ss addsimps [domain_overwrite]) 1);
by (fast_tac FOL_cs 1);
bw overwrite_def;

be UnE 1;
be CollectE 1;
br UnI1 1;
br CollectI 1;
br UnI2 1;
ba 1;
ba 1;
br UnI2 1;
ba 1;

val overwrite_assoc = result();


val overwrite_assoc_equality =
   prove_goal thy "[| g @@ h = k; f @@ k = l |] ==> (f @@ g) @@ h = l"
   (fn [p1,p2] => 
      [cut_facts_tac [p1, p2] 1,
                 asm_simp_tac (ZF_ss addsimps [overwrite_assoc]) 1]);

val prems =
  goalw thy [overwrite_def] "<x, b>:g ==> cons(<x, a>,f) @@ g = f @@ g";
by (cut_facts_tac prems 1);

br equalityI 1;
br subsetI 1;
be UnE 1;
be CollectE 1;
be consE 1;
by (hyp_subst_tac 1);
be swap 1;
by (simp_tac ZF_ss 1);
br domainI 1;
ba 1;
br UnI1 1;
br CollectI 1;
ba 1;
ba 1;
br UnI2 1;
ba 1; (* 17 *)

br subsetI 1;
be UnE 1;
be CollectE 1;
br UnI1 1;
br CollectI 1;
by (cases_tac "xa = <x, a>");
br consI2 1;
ba 1;
by (hyp_subst_tac 1);
br consI1 1;
ba 1;
br UnI2 1;
ba 1;
val cons_overwrite1 = result ();

val prems =
  goalw thy [overwrite_def] 
	"~x:domain(g) ==> cons(<x, a>,f) @@ g = cons(<x, a>, f @@ g)";
by (cut_facts_tac prems 1);

br equalityI 1;
br subsetI 1;
be UnE 1;
be CollectE 1;
be consE 1;
by (hyp_subst_tac 1);
br consI1 1;
br consI2 1;
br UnI1 1;
br CollectI 1;
ba 1;
ba 1;
br consI2 1;
br UnI2 1;
ba 1; (* 16 *)

br subsetI 1;
be consE 1;
by (hyp_subst_tac 1);
br UnI1 1;
br CollectI 1;
br consI1 1;
by (asm_simp_tac ZF_ss 1);
be UnE 1;
be CollectE 1;
br UnI1 1;
br CollectI 1;
br consI2 1;
ba 1;
ba 1;
br UnI2 1;
ba 1;
val cons_overwrite2 = result ();

val tacfun = fn prems => 
  [cut_facts_tac prems 1, 
   asm_simp_tac (ZF_ss addsimps[cons_overwrite1, cons_overwrite2]) 1];


val cons_overwrite_equality1 = prove_goal thy
   "[| <x,b> : g; f @@ g = h |] ==> cons (<x,a>, f) @@ g = h" tacfun;

val cons_overwrite_equality2 = prove_goal thy
   " [|  ~x : domain(g);  f @@ g = h; cons (<x,a>, h) = h' |] \
  \ ==> cons (<x,a>, f) @@ g = h'" tacfun;

